"""Programmatic creation of scanners through RegularGrammar objects.

The RegularGrammar object represents a group of formal regular expressions
which can be programatically transformed into a minimal DFA (Deterministic
Finite Automata) recognizing the given input pattern(s). This is done through a
series of transformations on the input taking it from an expression to an
epsilon NFA and finally to a minimal DFA.
"""
from uuid import uuid4
from string import printable
from copy import deepcopy


class RegularGrammar(object):
    """The RegularGrammar object responsible for creating the minimal DFA.

    RegularGrammar represents a collection of named formal regular expressions
    which can be programatically compiled into a minmal DFA capable of
    recognizing the input patterns. The resulting DFA is encoded as a table with
    a total delta (transition) function. Once created the object cannot be
    mutated, and it will remain static for the rest of it's lifetime. All
    information to properly understand and consume the DFA table can be queried
    through the exposed API functions.
    """

    # Map internal representation of operators -> literals
    _literals = dict(enumerate('*|.+?()[]'))

    # Map operator literals -> internal representation
    _operators = {v:k for k, v in _literals.items()}

    # Set of acceptable input characters (printable ascii) including:
    # uppercase, lowercase, digits, punctuation, whitespace. needed for
    # character class/range negation
    _characters = set(printable)

    # Operator precedence for the shunting yard algorithm.
    # (higher binds tighter)
    _precedence = {
        _operators['(']: (3, None),
        _operators[')']: (3, None),
        _operators['+']: (2, False),  # right-associative
        _operators['*']: (2, False),  # right-associative
        _operators['?']: (2, False),  # right-associative
        _operators['.']: (1, True),   # left-associative
        _operators['|']: (0, True),   # left-associative
    }

    def __init__(self, name, expressions):
        """Construct a scanner DFA given the input regular expressions.

        Attempt to initialize a RegularGrammar object with the specified name,
        recognizing the given named expression(s). Regular expressions must be
        specified following these guidelines:

           * only printable ascii characters (uppercase, lowercase, digits,
             punctuation, whitespace) are supported.
           * supported core operators (and extensions) include:
               * '|'    (union -> choice -> either or)
               * '.'    (concatenation -> combine)
               * '?'    (question -> choice -> 1 or none)
               * '*'    (kleene star -> repitition >= 0)
               * '+'    (kleene plus -> repitition >= 1)
               * ()     (grouping -> disambiguation -> any expression)
               * [ab]   (character class -> choice -> any specified character)
               * [a-c]  (character range -> choice -> any char between the two)
               * [^ab]  (character negation -> choice -> all but the specified)
           * other things to keep in mind (potential gotcha's):
               * character classes and ranges can be combined in the same set
                 of brackets, possibly multiple times (e.g. [abc-pqrs-z]).
               * character ranges can be specified as forward or backwards
                 (e.g. [a-c] or [c-a]).
               * '^' is required to come first after the bracket for
                 negation to properly work. In fact, '^' directly following a
                 bracket is always interpreted as negation.
               * '^' may be used with character classes or ranges.
               * if '^' is alone in the brackets (e.g. [^]) it is translated
                 as any possible input character (i.e. a 'wildcard').
               * if a literal '-' if required in a character class or range it
                 must be specified last so as to not be interpreted as a range.
               * for literal right brackets inside character ranges or classes
                 it must be escaped.
               * concatenation can be either implicit or explicit in the
                 given input expression(s).
               * operator literals (i.e. '|.?*+()[]') can be specified through
                 escape sequences using a backslash (i.e. '\').

        The input is then taken through a series of transformations taking it
        from a regular expression to an epsilon NFA and finally to a minimal
        DFA. The final minimal DFA produced will be minimized with respect to
        unreachable and nondistinguishable states. It will also have a total
        delta (transition) function, possibly including an sink/error state if
        necessary. It will also include a dictionary mapping the named input
        expressions (i.e. types) to there corresponding final state(s) within
        the resulting DFA.

        Args:
          name (str): the name of the scanner.
          expressions (dict[str, str]): token name/type and there pattern(s).

        Raises:
          TypeError: if name is not a string
          ValueError: if name is empty
          TypeError: if expressions is not a dict
          ValueError: if expressions is empty
          TypeError: if token identifier/type is not a string
          ValueError: if token identifier/type is empty
          TypeError: if token pattern is not a string
          ValueError: if token pattern is empty
          ValueError: if an invalid escape sequence is encountered.
          ValueError: if an unrecognized escape sequence is encountered.
          ValueError: if an empty escape sequence is encountered.
          ValueError: if character range has no starting bracket.
          ValueError: if character range has no ending bracket.
          ValueError: if left parenthesis are unbalanaced.
          ValueError: if right parenthesis are unbalanaced.
          ValueError: if concatentation is supplied with an improper arguments
          ValueError: if alternation is supplied with an improper arguments
          ValueError: if kleene star is supplied with an improper arguments
          ValueError: if kleene plus is supplied with an improper arguments
          ValueError: if choice is supplied with an improper arguments
          ValueError: if the input expression is invalid
        """
        if not isinstance(name, str):
            raise TypeError('name must be a string')

        if not name:
            raise ValueError('name must be non empty')

        self._name = name

        if not isinstance(expressions, dict):
            raise TypeError('expressions must be a dict')

        if not expressions:
            raise ValueError('expressions must be non empty')

        self._expressions = {}

        nfa = []
        for identifier, pattern in expressions.items():
            if not isinstance(identifier, str):
                raise TypeError('token name/type must be a string')

            if not identifier:
                raise ValueError('token name/type must be non empty')

            if not isinstance(pattern, str):
                raise TypeError('token pattern must be a string')

            if not pattern:
                raise ValueError('token pattern must be non empty')

            self._expressions[identifier] = pattern

            pattern = RegularGrammar._scan(pattern)
            pattern = RegularGrammar._expand_char_class_range(pattern)
            pattern = RegularGrammar._expand_concat(pattern)
            pattern = RegularGrammar._shunt(pattern)

            nfa.append((RegularGrammar._nfa(identifier, pattern)))

        Q, V, T, S, F, G = RegularGrammar._dfa(*RegularGrammar._merge_nfa(nfa))
        Q, V, T, S, F, G = RegularGrammar._total(Q, V, T, S, F, G)
        Q, V, T, S, F, G = RegularGrammar._alpha(*RegularGrammar._hopcroft(Q, V, T, S, F, G))

        self._states = Q
        self._alphas = V
        self._deltas = T
        self._start = S
        self._finals = F
        self._types = G

    @property
    def name(self):
        """Query for the name of the scanner.

        A readonly property which copies the scanners's name to protect against
        user mutations.

        Return:
          str: The given input name of the scanner.
        """
        return deepcopy(self._name)

    @property
    def expressions(self):
        """Query for the patterns recognized by the scanner.

        A readonly property which copies the scanners's token name/expression
        pairs to protect against user mutations.

        Return:
          list[tuple[str, str]]: the token name/expression pairs
        """
        return deepcopy(self._expressions)

    @property
    def states(self):
        """Query for states in the grammars equivalent minimal DFA.

        A readonly property which copies the scanner's states to protect against
        user mutation.

        Return:
          set[str]: all possible states in the resulting DFA.
        """
        return deepcopy(self._states)

    @property
    def alphabet(self):
        """Query for alphabet of characters recognized by the grammars DFA.

        A readonly property which copies the scannner's alphabet to protect
        against user mutation.

        Return:
          set[str]: all possible input characters to the resulting DFA.
        """
        return deepcopy(self._alphas)

    @property
    def transitions(self):
        """Query for the state transitions defining the grammars DFA.

        A readonly property which copies the scannner's transitions to protect
        against user mutation.

        Return:
          tuple[dict[str, int], dict[str, int], list[list[str]]]: the DFA
            encoded as a table.
        """
        return deepcopy(self._deltas)

    @property
    def start(self):
        """Query for the start state of the grammars DFA.

        A readonly property which copies the scanner's start state to protect
        against user manipulation.

        Return:
          str: the start state of the grammars DFA.
        """
        return deepcopy(self._start)

    @property
    def accepting(self):
        """Query for all accepting states of the grammars DFA.

        A readonly property which copies the scanner's accepting states to
        protect against user manipulation.

        Return:
          set[str]: the resulting DFA's accepting states.
        """
        return deepcopy(self._finals)

    @property
    def types(self):
        """Query for the dictionary labeling all types to ther final state(s).

        A readonly property which copies the scanner's types to protect against
        user manipulation.

        Return:
          dict[str, set[str]]: the resulting DFA's token types.
        """
        return deepcopy(self._types)

    @staticmethod
    def _scan(expr):
        """transform the external representation to an internal one.

        Convert an external representation of a token (regular expression) to
        an internal one. Ensures all characters and escape sequences are valid.

        Args:
          expr (str): an external representation of a regular expression.

        Return:
          list[str, int]: an internal representation of a regular expression.

        Raises:
          ValueError: if an invalid escape sequence is encountered.
          ValueError: if an unrecognized escape sequence is encountered.
          ValueError: if an empty escape sequence is encountered.
        """
        output = []
        escape = False
        for char in expr:
            if escape:
                escape = False
                if char in RegularGrammar._operators:
                    output.append(char)
                elif char == '\\':
                    output.append('\\')
                elif char == 's':
                    output.append(' ')
                elif char == 't':
                    output.append('\t')
                elif char == 'n':
                    output.append('\n')
                elif char == 'r':
                    output.append('\r')
                elif char == 'f':
                    output.append('\f')
                elif char == 'v':
                    output.append('\v')
                else:
                    raise ValueError('Error: invalid escape seq: \\' + char)
            else:
                if char == '\\':
                    escape = True
                elif char in RegularGrammar._operators:
                    output.append(RegularGrammar._operators[char])
                elif char in RegularGrammar._characters:
                    output.append(char)
                else:
                    raise ValueError('Error: unrecognized character: ' + char)
        if escape:
            raise ValueError('Error: empty escape sequence')
        return output

    @staticmethod
    def _expand_char_class_range(expr):
        """expand any character classes/ranges present in the expression.

        Expand the internal representation of the expression so that
        character classes and ranges are eliminated.

        Args:
          expr (list[str, int]): an internal representation of a regular
              expression possible with character ranges and/or classes.

        Return:
          list[str, int]: an internal representation of a regular expression
              with all character classes and ranges elimintated.

        Raises:
          ValueError: if character range has no starting bracket.
          ValueError: if character range has no ending bracket.
        """
        output, literals = [], []
        expansion, negation, prange = False, False, False
        for char in expr:
            if char == RegularGrammar._operators['['] and not expansion:
                expansion = True
            elif char == RegularGrammar._operators[']']:
                if not expansion:
                    raise ValueError('Error: Invalid character class/range; no start')
                expansion = False
                if prange:
                    prange = False
                    literals.append('-')
                if negation:
                    negation = False
                    literals = list(RegularGrammar._characters - set(literals))
                if literals:
                    literals = list(set(literals))
                    output.append(RegularGrammar._operators['('])
                    while literals:
                        output.append(literals.pop())
                        output.append(RegularGrammar._operators['|'])
                    output[-1] = RegularGrammar._operators[')']
            elif not expansion:
                output.append(char)
            elif char == '^' and not literals and not negation:
                negation = True
            elif char == '-' and literals and not prange:
                prange = True
            elif prange:
                prange = False
                _char = literals.pop()
                literals.extend(map(chr, range(ord(min(_char, char)), ord(max(_char, char))+1)))
            else:
                literals.append(RegularGrammar._literals.get(char, char))
        if expansion:
            raise ValueError('Error: character class/range end not specified')
        return output

    @staticmethod
    def _expand_concat(expr):
        """make concatenation explicit throughout the expression.

        Expand the internal representation of the expression so that
        concatentation is explicit throughout.

        Args:
          expr (list[str, int]): an internal representation of a regular
              expression possibly using implicit concatenation.

        Return:
          list[str, int]: an internal representation of a regular expression
              with concatenation explicit throughout.
        """
        if not expr:
            return expr

        output = [expr[0]]
        for elem in expr[1:]:
            if output[-1] != RegularGrammar._operators['('] and \
               output[-1] != RegularGrammar._operators['|'] and \
               output[-1] != RegularGrammar._operators['.'] and \
               (elem == RegularGrammar._operators['('] or
                elem not in RegularGrammar._literals):
                output.append(RegularGrammar._operators['.'])
            output.append(elem)
        return output

    @staticmethod
    def _shunt(expr):
        """Convert infix regular expression notation to postfix.

        Convert the input expression to be entirely in postfix notation (RPN;
        Reverse Polish Notation) allowing all parenthesis to be dropped.
        Adapted from Dijkstra's Shunting yard algorithm which can be viewed
        @https://en.wikipedia.org/wiki/Shunting-yard_algorithm.

        Args:
          expr (list[str, int]): an infix regular expression.

        Return:
          list[str, int]: a postfix regular expression.

        Raises:
          ValueError: if left parenthesis are unbalanaced.
          ValueError: if right parenthesis are unbalanaced.
        """
        stack, queue = [], []  # operators, output expression

        for token in expr:
            if token == RegularGrammar._operators['(']:
                stack.append(RegularGrammar._operators['('])
            elif token == RegularGrammar._operators[')']:
                while stack and stack[-1] != RegularGrammar._operators['(']:
                    queue.append(stack.pop())
                if not stack:
                    raise ValueError('Error: unbalanced right parenthesis')
                stack.pop()
            elif token in RegularGrammar._precedence:
                while stack and stack[-1] != RegularGrammar._operators['('] and\
                      RegularGrammar._precedence[token][0] <= \
                      RegularGrammar._precedence[stack[-1]][0]\
                      and RegularGrammar._precedence[token][1]:  # left-associative?
                    queue.append(stack.pop())
                stack.append(token)
            else:  # it's a character
                queue.append(token)

        while stack:
            token = stack.pop()
            if token == RegularGrammar._operators['(']:
                raise ValueError('Error: unbalanced left parenthesis')
            queue.append(token)

        return queue

    @staticmethod
    def _state():
        """Generate a new universally unique state label.

        The state label is a string prepresentation of a UUID type 4.

        Return:
          str: a unique state name
        """
        return str(uuid4())

    @staticmethod
    def _nfa(name, expr):
        """convert a postfix notation regular expression to an epsilon NFA.

        Attempt to convert an internal representation of a regular expression
        in RPN to an epsilon NFA. Operators handled: union |, kleene star *,
        concatenation ., literals, and syntax extensions kleene plus + and
        choice ?. Adapted to a iterative stacked based evaluation algorithm
        (standard RPN evaluation algorithm) from thompson construction as
        described in section 4.1 in 'A taxonomy of finite automata construction
        algorithms' by Bruce Watson,
        located @http://alexandria.tue.nl/extra1/wskrap/publichtml/9313452.pdf

        Args:
          name (str): the identifier/type of the expression.
          expr (list[str, int]): a regular expression in postfix notation.

        Return:
          set[str]: the set of states, Q
          set[str]: the set of alphabet characters, V
          set[tuple[str, str, str]]: the set of state transitions, T
          dict[str, set[str]]: the set of epsilon transitions, E
          str: the start state, S
          str: the final state, F
          dict[str, str]: map of identifier/type to final state, G

        Raises:
          ValueError: if concatentation is supplied with an improper arguments
          ValueError: if alternation is supplied with an improper arguments
          ValueError: if kleene star is supplied with an improper arguments
          ValueError: if kleene plus is supplied with an improper arguments
          ValueError: if choice is supplied with an improper arguments
          ValueError: if the input expression is invalid
        """
        Q = set()   # states
        V = set()   # input symbols (alphabet)
        T = set()   # transition relation: T in P(Q x V x Q)
        E = dict()  # e-transition relation: E in P(Q x Q)
        S = None    # start state S in Q
        F = None    # accepting state F in Q
        G = dict()  # map type to the final state(s)

        def e_update(start, final):
            """An internal helper to update the epsilon dictionary cache."""
            E[start] = E.get(start, set())
            E[start].add(final)

        stk = []  # NFA machine stk
        for token in expr:
            if token == RegularGrammar._operators['.']:
                if len(stk) < 2:
                    raise ValueError('Error: not enough args to op .')
                p, F = stk.pop()
                S, q = stk.pop()
                e_update(q, p)
            elif token == RegularGrammar._operators['|']:
                if len(stk) < 2:
                    raise ValueError('Error: not enough args to op |')
                p, q = stk.pop()
                r, t = stk.pop()
                S, F = RegularGrammar._state(), RegularGrammar._state()
                e_update(S, p)
                e_update(S, r)
                e_update(q, F)
                e_update(t, F)
            elif token == RegularGrammar._operators['*']:
                if len(stk) < 1:
                    raise ValueError('Error: not enough args to op *')
                p, q = stk.pop()
                S, F = RegularGrammar._state(), RegularGrammar._state()
                e_update(S, p)
                e_update(q, p)
                e_update(q, F)
                e_update(S, F)
            elif token == RegularGrammar._operators['+']:
                if len(stk) < 1:
                    raise ValueError('Error: not enough args to op +')
                p, q = stk.pop()
                S, F = RegularGrammar._state(), RegularGrammar._state()
                e_update(S, p)
                e_update(q, p)
                e_update(q, F)
            elif token == RegularGrammar._operators['?']:
                if len(stk) < 1:
                    raise ValueError('Error: not enough args to op ?')
                p, q = stk.pop()
                S, F = RegularGrammar._state(), RegularGrammar._state()
                e_update(S, p)
                e_update(S, F)
                e_update(q, F)
            else:  # it's a character
                S, F = RegularGrammar._state(), RegularGrammar._state()
                V.add(token)
                T.add((S, token, F))
            Q.update([S, F])
            stk.append((S, F))

        if len(stk) != 1:
            raise ValueError('Error: invalid expression')
        S, F = stk.pop()
        G[name] = F
        return Q, V, T, E, S, F, G

    @staticmethod
    def _merge_nfa(nfa):
        """merge a list of NFAs into a single NFA.

        Merge multiple NFAs into a single NFA with a new start state containing
        epsilon transitions to each individual machine's start state.

        Args:
          nfa: list[tuple[
                          set[str],
                          set[str],
                          set[tuple[str, str, str]],
                          dict[str, set[str]],
                          str,
                          str,
                          dict[str, str]
                    ]]: the input NFAs to merge together.

        Return:
          set[str]: the merged set of states, Q
          set[str]: the merged set of alphabet characters, V
          set[tuple[str, str, str]]: the merged set of state transitions, T
          dict[str, set[str]]: the merged set of epsilon transitions, E
          str: the new start state, S
          set[str]: the merged set of final states, F
          dict[str, str]: merged mapping of expression type to final state(s), G
        """
        S = RegularGrammar._state()
        Q, V, T, E, S, F, G = set(), set(), set(), dict(), S, set(), dict()
        E[S] = set()
        for _nfa in nfa:
            Q.update(_nfa[0])
            V.update(_nfa[1])
            T.update(_nfa[2])
            E[S].add(_nfa[4])
            for state, etransitions in _nfa[3].items():
                E[state] = E.get(state, set()) | etransitions
            F.add(_nfa[5])
            for name, state in _nfa[6].items():
                G[name] = state
        return Q, V, T, E, S, F, G

    @staticmethod
    def _e_closure(q, E, cache):
        """find the epsilon closure of the current state and cache the result.

        Find the epsilon closure of a given state given all epsilon transitions.
        A cache is utilized to speed things up for repeated invocations. Stated
        in set notation: { q' | q ->*e q' }, from a given start state q find all
        states q' which are reachable using only epsilon transitions, handling
        cycles appropriately.

        Args:
          q (str): the state to find the e-closure of.
          E (dict[str, set[str]]): the set of e-transitions.
          cache (dict[str, set[str]]): previously computed e-closures.

        Return:
          set[str]: the e-closure of state q.
        """
        if q in cache:
            return cache[q]

        cache[q] = closure = set()
        explore = set([q])
        while explore:
            q = explore.pop()
            if q not in closure:
                closure.add(q)
                # perform a single step: { q' | q ->e q' }
                explore.update(E.get(q, set()))

        return closure

    @staticmethod
    def _dfa(Q, V, T, E, S, F, G):
        """convert the input NFA (possibly with epsilon transitions) to a DFA.

        Convert the epsilon NFA to a DFA using subset construction and
        e-closure conversion. Only states wich are reachable from the start
        state are considered. This results in a minimized DFA with reguard to
        reachable states, but not with reguard to nondistinguishable states.

        Args:
          Q (set[str]): set of NFA states
          V (set[str]): set of NFA alphabet characters
          T (set[tuple[str, str, str]]): set of state transitions
          E (dict[str, set[str]]): set of epsilon transitions
          S (str): start state
          F (set[str]): set of final states
          G (dict[str, str]): mapping of expressions types to final states

        Return:
          set[frozenset[str]]: set of DFA states
          set[str]: set of characters in the DFA alphabet
          set[tuple[frozenset[str], str, frozenset[str]]]: the DFA transitions
          frozenset[str]: the DFA start state
          set[frozenset[str]]: set of DFA final states
          dict[str, set[frozenset[str]]]: mapping of types to final states
        """
        cache, Gp = dict(), dict()
        Sp = frozenset(RegularGrammar._e_closure(S, E, cache))
        Qp, Fp, Tp, explore = set(), set(), set(), set([Sp])
        while explore:
            in_state = explore.pop()  # DFA state; set of NFA states
            if in_state not in Qp:
                Qp.add(in_state)
                if F & in_state:
                    Fp.add(in_state)
                qps = {}
                for t in T:
                    if t[0] in in_state:
                        out_states = qps[t[1]] = qps.get(t[1], set())
                        out_states.update(RegularGrammar._e_closure(t[2], E, cache))
                for alpha, out_state in qps.items():
                    out_state = frozenset(out_state)
                    explore.add(out_state)
                    Tp.add((in_state, alpha, out_state))

        for name, nfa_final in G.items():
            for dfa_final in Fp:
                if nfa_final in dfa_final:
                    Gp[name] = Gp.get(name, set()) | set([dfa_final])

        return Qp, V, Tp, Sp, Fp, Gp

    @staticmethod
    def _total(Q, V, T, S, F, G):
        """extend the DFA's transition function, making it total.

        Make the DFA's delta function total, if not already, by adding a
        sink/error state. All unspecified state transitions are then specified
        by adding a transition to the new sink/error state. A new entry is also
        made into G to track this new sink/error type which is accessible as
        '_sink'.

        Args:
          Q (set[frozenset[str]]): set of DFA states
          V (set[str]): the DFA alphabet
          T (set[tuple[frozenset[str], str, frozenset[str]]]): DFA transitions
          S (frozenset[str]): the DFA start state
          F (set[frozenset[str]]): the set of DFA final states
          G (dict[str, set[frozenset[str]]]): DFA pattern name to final state(s)

        Return:
          set[frozenset[str]]: possibly extended set of DFA states
          set[str]: the given input alphabet
          tuple[
                dict[frozenset[str], int],
                dict[str, int],
                list[list[frozenset[str]]]
               ]: a possibly extended transition function converted to a table
          frozenset[str]: the input start state
          set[frozenset[str]]: the input final state(s)
          dict[str, set[frozenset[str]]]: the type to pattern mapping
        """
        q_err = frozenset([RegularGrammar._state()])
        if len(T) != len(Q) * len(V):
            Q.add(q_err)
            G['_sink'] = set([q_err])

        states = {v:k for k, v in enumerate(Q)}
        symbols = {v:k for k, v in enumerate(V)}
        table = [[q_err for _ in states] for _ in symbols]
        for (state, symbol, dest) in T:
            table[symbols[symbol]][states[state]] = dest

        return Q, V, (states, symbols, table), S, F, G

    @staticmethod
    def _hopcroft(Q, V, T, S, F, G):
        """reduce the DFA state complexity by merging nondistinguishable states.

        Minimize the DFA with reguard to nondistinguishable states using
        hopcrafts algorithm, which merges states together based on partition
        refinement.

        Args:
          Q (set[frozenset[str]]): the set of DFA states
          V (set[str]): the set of DFA input characters
          T (tuple[
                   dict[frozenset[str], int],
                   dict[str, int],
                   list[list[frozenset[str]]]
                  ]): the DFA transition function as a table
          S (frozenset[str]): the DFA start state
          F (set[frozenset[str]]): the set of DFA final states
          G (dict[str, set[frozenset[str]]]): map of type to DFA final state(s)

        Return:
          set[set[frozenset[frozenset[str]]]]: set of possible merged states
          set[str]: the given set of input chacters
          tuple[
                dict[set[frozenset[frozenset[str]]], int],
                dict[str, int],
                list[list[set[frozenset[frozenset[str]]]]]
               ]: possibly updated and reduced table with merged states
          set[frozenset[frozenset[str]]]: a possibly updated start state.
          set[set[frozenset[frozenset[str]]]]: a possibly updated set of final
              states.
          dict[str, set[set[frozenset[frozenset[str]]]]: a possibly updated map
              of token to final state(s).
        """
        (states, symbols, T) = T
        Q, F = frozenset(Q), frozenset(F)

        partitions = set([F, Q - F]) - set([frozenset()])  # if Q - F was empty
        explore = set([F])

        while explore:
            selection = explore.pop()
            for v_idx in symbols.values():
                _selection = {q for q, q_idx in states.items() if T[v_idx][q_idx] in selection}
                _selection = frozenset(_selection)
                _partitions = set()
                for partition in partitions:
                    split_1 = _selection & partition
                    split_2 = partition - _selection
                    if split_1 and split_2:
                        _partitions.update([split_1, split_2])
                        if partition  in explore:
                            explore.remove(partition)
                            explore.update([split_1, split_2])
                        elif len(split_1) <= len(split_2):
                            explore.update([split_1])
                        else:
                            explore.update([split_2])
                    else:
                        _partitions.add(partition)
                partitions = _partitions

        _states = dict(zip(partitions, range(len(partitions))))
        Tp = [[None for _ in partitions] for symbol in V]
        for source in states:
            for symbol in V:
                dest = T[symbols[symbol]][states[source]]
                state_1, state_2 = None, None
                for part in partitions:
                    if source in part:
                        state_1 = part
                        if state_2:
                            break
                    if dest in part:
                        state_2 = part
                        if state_1:
                            break
                Tp[symbols[symbol]][_states[state_1]] = state_2

        Sp = None
        for part in partitions:
            if S in part:
                Sp = part
                break

        Fp = {partition for partition in partitions if partition & F}

        Gp = dict()

        for name, dfa_finals in G.items():
            if name == '_sink': # special case (i.e not a final state)
                Gp[name] = {part for part in partitions if part & G['_sink']}
                continue
            for dfa_final in dfa_finals:
                for dfa_merged_final in Fp:
                    if dfa_final in dfa_merged_final:
                        Gp[name] = Gp.get(name, set()) | set([dfa_merged_final])

        return partitions, V, (_states, symbols, Tp), Sp, Fp, Gp

    @staticmethod
    def _alpha(Q, V, T, S, F, G):
        """rename/convert states for better legibility.

        Perform an alpha rename on all DFA states to simplify the
        representation which the end user will consume.

        Args:
          Q (set[set[frozenset[frozenset[str]]]]): the set of states
          V (set[str]): the input alphabet chacters
          T (tuple[
                   dict[set[frozenset[frozenset[str]]], int],
                   dict[str,  int],
                   list[list[set[frozenset[frozenset[str]]]]]
                  ]): the delta function as a table
          S (set[frozenset[frozenset[str]]]): the start state
          F (set[set[frozenset[frozenset[str]]]]): the set of final states
          G (dict[str, set[set[frozenset[frozenset[str]]]]]): the type to state
              mapping.

        Return:
          set[str]: the renamed start states
          set[str]: the input alphabet
          tuple[dict[str, int], dict[str, int], list[list[str]]]: the updated
              transition function
          str: the updated start state
          set[str]: the updated finish state(s)
          dict[str, set[str]]: the updated token to final state mapping.
        """
        rename = {q: RegularGrammar._state() for q in Q}
        Qp = set(rename.values())
        (states, symbols, table) = T
        states = {rename[state]:idx for state, idx in states.items()}
        Tp = (states, symbols, [[rename[col] for col in row] for row in table])
        Sp = rename[S]
        Fp = {rename[f] for f in F}
        Gp = {g:set([rename[s] for s in states]) for g, states in G.items()}
        return Qp, V, Tp, Sp, Fp, Gp



"""C generator.

A scanner/parser generator targeting c. Generates header (.h) and source (.c)
files.
"""
import datetime
from spag.generator import Generator


class C(Generator):
    """C generator.

    A simple object for compiling scanner's and/or parser's to c programs.
    """

    _reserved = {
        "auto", "break", "case", "char", "const", "continue", "default", "do",
        "int", "long", "register", "return", "short", "signed", "sizeof",
        "static", "struct", "switch", "typedef", "union", "unsigned", "void",
        "volatile", "while", "double", "else", "enum", "extern", "float", "for",
        "goto", "if"
    }

    def _sanatize(self, name):
        """
        Sanatize the name so it is safe for c compilation following the rules:
         1. Characters other than a-z, A-Z, 0-9, or _ become '_'
         2. Names beginning with a number will be prefixed with a '_'
         3. Reserved c keyword will be prefixed with a '_'
        """
        _name = ''
        for char in name:
            if char.isalnum():
                _name += char
            else:
                _name += '_'

        if _name in self._reserved or name[0].isdigit():
            _name = '_' + _name

        return _name

    def _generate_file_header(self, filename, author, source, message, libs):
        time = datetime.datetime.utcnow().isoformat("T") + "Z"
        libs = ['#include <{0}.h>'.format(lib) for lib in libs]
        return """\
/******************************************************************************
 * File:    {0: <66}*
 * Author:  {1: <66}*
 * Created: {2: <66}*
 * Archive: {3: <66}*
 *                                                                            *
 *{4: ^76}*
 ******************************************************************************/

{6}{5}

""".format(filename, author, time, source, message,
           '\n'.join(libs), self._generate_section_header("imports"))

    @staticmethod
    def _generate_section_header(name):
        return """\
/******************************************************************************
 *{0: ^76}*
 ******************************************************************************/
""".format(name.upper())


    def _generate_token_api(self, name):
        types = []
        for token_name, pattern in self.scanner.expressions.items():
            types.append('  {0: <23} // {1}'.format(token_name.upper()+",", pattern))
        return """\
{2}
// Token's abstract over the character input stream.
typedef struct {0}_token {0}_token_t;

// Token's are associated to one of the types below.
typedef enum {{
{1}
}} {0}_token_type_t;

// Attempt construction of a new {0} token.
{0}_token_t *new_{0}_token({0}_token_type_t type,
                           char *text,
                           unsigned long text_len,
                           char *source,
                           unsigned long source_len,
                           unsigned long line,
                           unsigned long column);

// Free a token constructed using new_{0}_token.
void free_{0}_token({0}_token_t *{0}_token);

// Query for the tokens associated type.
{0}_token_type_t type({0}_token_t *{0}_token);

// Query for the string representation of the token.
char *text({0}_token_t *{0}_token);

// Query for the file in which the token was read.
char *source({0}_token_t *{0}_token);

// Query for the starting line on which the token was read.
unsigned long line({0}_token_t *{0}_token);

// Query for the starting column on which the token was read.
unsigned long column({0}_token_t *{0}_token);

""".format(name, "\n".join(types), self._generate_section_header("tokens")), """\
{1}
typedef struct {0}_token {{
  {0}_token_type_t type;
  char *text;
  char *source;
  unsigned long line;
  unsigned long column;
}} {0}_token_t;

{0}_token_t *new_{0}_token({0}_token_type_t type,
                           char *text,
                           unsigned long text_len,
                           char *source,
                           unsigned long source_len,
                           unsigned long line,
                           unsigned long column) {{
  {0}_token_t *{0}_token;
  if(!(({0}_token = calloc(1, sizeof({0}_token_t))) &&
       ({0}_token->text = malloc(sizeof(char)*(text_len+1))) &&
       ({0}_token->source = malloc(sizeof(char)*(source_len+1))))) {{
    free({0}_token->source);
    free({0}_token->text);
    free({0}_token);
    return NULL;
  }}

  {0}_token->type = type;
  {0}_token->line = line;
  {0}_token->column = column;
  strncpy({0}_token->text, text, sizeof(char)*text_len);
  {0}_token->text[text_len] = 0;
  strncpy({0}_token->source, source, sizeof(char)*source_len);
  {0}_token->source[source_len] = 0;

  return {0}_token;
}}

void free_{0}_token({0}_token_t *{0}_token) {{
  free({0}_token->text);
  free({0}_token->source);
  free({0}_token);
}}

char *text({0}_token_t *{0}_token) {{ return {0}_token->text; }}

char *source({0}_token_t *{0}_token) {{ return {0}_token->source; }}

unsigned long line({0}_token_t *{0}_token) {{ return {0}_token->line; }}

unsigned long column({0}_token_t *{0}_token) {{ return {0}_token->column; }}

{0}_token_type_t type({0}_token_t *{0}_token) {{ return {0}_token->type; }}

""".format(name, self._generate_section_header("tokens"))

    def _encode_dfa(self, name):
        state, symbol, T = self.scanner.transitions
        final_states = self.scanner.accepting
        expressions = self.scanner.expressions
        types = self.scanner.types

        labels, label = {}, 0
        for state_id in state.keys():
            label += 1
            labels[state_id] = "L{0}".format(label)

        program = """\
  goto {0};

""".format(labels[self.scanner.start])

        for in_state, state_key in state.items():
            if in_state in types.get('_sink', set()):
                continue  # NOTE: Don't encode error state explicitly.
            fallthrough = dict()
            for char, sym_key in symbol.items():
                _char = ord(char)
                if _char < 0 or _char > 255:
                    raise ValueError("Invalid Input: encountered non ascii character\n")
                end_state = T[sym_key][state_key]
                if end_state in fallthrough:
                    fallthrough[end_state].append(char)
                else:
                    fallthrough[end_state] = [char]

            cases = ""
            for end_state, char_list in fallthrough.items():
                if end_state in types.get('_sink', set()):
                    continue  # NOTE: Don't encode error transitions explicitly.
                char_list = [(hex(ord(char)), repr(char)) for char in char_list]
                for hex_repr, char_repr in sorted(char_list, key=lambda x: chr(int(x[0], base=16))):
                    cases += """\
    case {0: <5} // {1}
""".format("{0}:".format(hex_repr), char_repr)
                if end_state in final_states:
                    _type = None
                    for pname, _ in expressions.items():
                        if end_state in types[pname]:
                            _type = pname
                            break
                    cases += """\
      {0}_scanner->last_final_pos = ftell({0}_scanner->input);
      {0}_scanner->last_final_type = {1};
      goto {2};
""".format(name, _type.upper(), labels[end_state])
                else:
                    cases += """\
      goto {0};
""".format(labels[end_state])
            # NOTE: Uses 'default' as catch all to revert to last valid final state.
            # This implementes maximal munch. If not final state was found an
            # error is returned which means 1 of 2 possibilities: unrecognized or invalid input.
            program += """\
{0}:
  if(feof({1}_scanner->input)) {{ return {1}_EOF; }}

  switch({1}_peek({1}_scanner)) {{
{2}    default:
      if({1}_scanner->last_final_pos) {{
        fprintf(stdout, "TYPE: %d @%ld\\n", {1}_scanner->last_final_type, {1}_scanner->last_final_pos);
        return {1}_read_token({1}_scanner, {1}_scanner->last_final_type);
      }}
      return {1}_INVALID_INPUT;
  }}

""".format(labels[in_state], name, cases)

        return program

    # NOTE:
    # - switch to internal buffer, allowing for pipes.
        # -> more scanner state and a buffer API?
        # - NFA to be updated with types in final state for uniqueness when ->dfa->hopcroft??
        #   - cannot be done. example is most langs do relop -> < | > | <= | >= ...
    def _generate_scanner_api(self, name):
        return """\
{1}
// Abstract over the reading of {0}_token_t's.
typedef struct {0}_scanner {0}_scanner_t;

// Attempt the creation of a new scanner given the path to a file.
{0}_scanner_t *new_{0}_scanner(char *fpath);

// Free the scanner and close the associated file.
void free_{0}_scanner({0}_scanner_t *{0}_scanner);

// Return most recently read token of the given scanner.
{0}_token_t *{0}_token({0}_scanner_t *{0}_scanner);

// Errors associated with scanning.
typedef enum {{
  {0}_NIL,                 // no error.
  {0}_EOF,                 // end of file.
  {0}_INVALID_INPUT,       // input not recognized by scanner.
  {0}_OUT_OF_MEMORY,       // failed to allocate memory.
  {0}_UNKNOWN_ERROR,       // error not known.
}} {0}_scan_error_t;

// Attempt to scan a token from the file. 1 if successful, otherwise 0.
// If failure occurs the token will still contain the relevant details of the
// unrecognized token except for its type.
{0}_scan_error_t {0}_scan({0}_scanner_t *{0}_scanner);
""".format(name, self._generate_section_header("scanner")), """\
{2}
typedef struct {0}_scanner {{
  FILE *input;
  int peek;
  long int offset;
  char *text;
  char *source;
  unsigned int length;
  unsigned int line;
  unsigned int column;
  {0}_token_t *token;
  long int last_final_pos;
  {0}_token_type_t last_final_type;
}} {0}_scanner_t;

{0}_scanner_t *new_{0}_scanner(char *fpath) {{
  FILE *f = fopen(fpath, "r");
  if(!f) {{ return NULL; }}

  {0}_scanner_t *{0}_scanner;
  if(!({0}_scanner = calloc(1, sizeof({0}_scanner_t))) &&
      ({0}_scanner->token = malloc(sizeof({0}_token_t)))) {{
    free({0}_scanner->token);
    free({0}_scanner);
    return NULL;
  }}

  {0}_scanner->input = f;
  //{0}_scanner->peek = fgetc(f);
  {0}_scanner->offset = 0;
  {0}_scanner->text = NULL;
  {0}_scanner->source = fpath;
  {0}_scanner->length = 0;
  {0}_scanner->line = 0;
  {0}_scanner->column = 0;
  {0}_scanner->last_final_pos = -1;
  //{0}_scanner->last_final_type = 0;

  return {0}_scanner;
}}

void free_{0}_scanner({0}_scanner_t *{0}_scanner) {{
  fclose({0}_scanner->input);
  free({0}_scanner->text);
  free({0}_scanner->token);
  free({0}_scanner);
}}

{0}_token_t *{0}_token({0}_scanner_t *{0}_scanner) {{
  return {0}_scanner->token;
}}

int {0}_peek({0}_scanner_t *{0}_scanner) {{
  {0}_scanner->peek = fgetc({0}_scanner->input);
  {0}_scanner->column++;
  if({0}_scanner->peek == '\\n') {{
    {0}_scanner->line++;
    {0}_scanner->column = 0;
  }}
  return {0}_scanner->peek;
}}

{0}_scan_error_t {0}_read_token({0}_scanner_t *{0}_scanner,
                                {0}_token_type_t {0}_type) {{
  {0}_scanner->length = ftell({0}_scanner->input) - {0}_scanner->offset;

  {0}_scanner->text = realloc({0}_scanner->text, sizeof(char)*{0}_scanner->length);
  if(!{0}_scanner->text) {{ return {0}_OUT_OF_MEMORY; }}

  fseek({0}_scanner->input, {0}_scanner->offset, SEEK_SET);
  if(fread({0}_scanner->text,
           sizeof(char),
           sizeof(char)*{0}_scanner->length,
           {0}_scanner->input) != sizeof(char)*{0}_scanner->length) {{
    return {0}_UNKNOWN_ERROR;
  }}

  if(!({0}_scanner->token = new_{0}_token({0}_type,
                                          {0}_scanner->text,
                                          {0}_scanner->length,
                                          {0}_scanner->source,
                                          0,
                                          {0}_scanner->line,
                                          {0}_scanner->column))) {{
    return {0}_OUT_OF_MEMORY;
  }}

  return {0}_NIL;
}}

{0}_scan_error_t {0}_scan({0}_scanner_t *{0}_scanner) {{
  {0}_scanner->offset = ftell({0}_scanner->input);

{1}}}
""".format(name, self._encode_dfa(name), self._generate_section_header("scanner"))

    def _generate_ast_api(self, name):
        # NOTE: define AST prototypes and defs
        _ = name
        _ = self
        ast_header, ast_source = "", ""
        return ast_header, ast_source

    def _encode_bnf(self, name):
        # NOTE: graph encoded as GOTOs;
        # automatically throw away tokens not in the bnf (whitespace, comments, etc)
        _ = name
        _ = self
        return ""
        #production_rules = ""
        #for nonterm, rule in self._parser.rules():
        #    production_rules += "// {0: <30} ::= {1}\n".format(nonterm, " ".join(rule))
        # """\
        # // Start production ::= {2}
        #
        # {3}
        #
        # // Attempt to parse into an AST of tokens. 1 if successful, otherwise 0.
        # int {0}(FILE *f);
        # """.format(parse_func, self._generate_section_header("::BNF GRAMMMAR::"),
        #            self._parser.start(), production_rules)

    def _generate_parser_api(self, name):
        # NOTE: define parser prototypes and defs
        parser_header, parser_source = "", """\
        {0} {1} {2}
        """.format(name, self._encode_bnf(name), self._generate_section_header("parser"))
        return parser_header, parser_source

    def _translate(self):
        """Override the superclass method to generate source code.

        Attempt to generate the c source(.c) and header(.h) files for the
        corresponding scanner and/or parser currently set in the object.
        """
        author = '**AUTO GENERATED**'
        source = 'https://github.com/rrozansk/Scanner-Parser-Generator'
        warning = 'WARNING!! AUTO GENERATED FILE, DO NOT EDIT!'
        libs = ['stdio']
        filename = self._sanatize(self.filename)
        if self.scanner:
            filename += self._sanatize('_'+self.scanner.name)
        if self.parser:
            filename += self._sanatize('_'+self.parser.name)

        header = self._generate_file_header(filename+".h",
                                            author,
                                            source,
                                            warning,
                                            libs)

        libs.extend(["stdlib", "string", filename])
        source = self._generate_file_header(filename+".c",
                                            author,
                                            source,
                                            warning,
                                            libs)

        if self.scanner is not None:
            scan_func = self._sanatize(self.scanner.name)
            token_header, token_source = self._generate_token_api(scan_func)
            scanner_header, scanner_source = self._generate_scanner_api(scan_func)

            header += token_header + scanner_header
            source += token_source + scanner_source


        if self.parser is not None:
            parse_func = self._sanatize(self.parser.name)
            ast_header, ast_source = self._generate_ast_api(parse_func)
            parser_header, parser_source = self._generate_parser_api(parse_func)

            header += ast_header + parser_header
            source += ast_source + parser_source

        header = """\
#ifndef {0}
#define {0}

{1}
#endif
""".format(filename, header)

        return {
            filename+'.h': header,
            filename+'.c': source,
        }



"""Golang generator.

A scanner/parser generator targeting go. Generates a single golang (.go) file.
"""
from spag.generator import Generator


class Go(Generator):
    """Golang generator.

    A simple object for compiling scanner's and/or parser's to golang programs.
    """

    def _translate(self):
        """Override the superclass method to generate source code.

        Attempt to generate the golang (.go) source file with the corresponding
        scanner and/or parser currently set in the object.
        """
        return {self.filename+'.go': super(Go, self)._translate()}



"""Base Generator which protects against bad input and return values.

The Generator is a simple superclass which protect subclasses from bad user
input when trying to generate output programs. Likewise, it also protects the
user from bad developer output. It does this by performing simple validations
before and after program generation is attempted. The base Generator also
controls the getting/setting of options used for code generation. Each targeted
language for generation should be implemented as a subclass which only needs to
override the '_translate(self)' method.
"""
from spag.parser import ContextFreeGrammar
from spag.scanner import RegularGrammar


class Generator(object):
    """The Base Generator which performs [in/out]put validations.

    A simple superclass object for performing validation on user input as well
    as developer output so subclasses can focus on translating the IR returned
    by the ContextFreeGrammar and RegularGrammar classes.
    """

    def __init__(self):
        """Construct a base Generator for code generation with default options.

        instantiate a new generator with the defaults populated for all the
        properties in the object.
        """
        self._scanner = None
        self._parser = None
        self._filename = 'out'
        self._encoding = 'direct'

    @property
    def scanner(self):
        """Get or set the scanner to be used for generation, if any.

        Query for or attempt to set the scanner to be used for code generation.
        This allows the child language Generators to be reused rather than
        recreated if attempting to generate source for multiple scanners.

        Args:
          scanner (RegularGrammar): The scanner to be used for generation. If
              generation of the scanner is no longer desired it may also be set
              to None.

        Return:
          None: The default if the scanner was not previously set.
          RegularGrammar: If scanner was set.

        Raises:
          TypeError: If `scanner` is not None or a RegularGrammar when
              attempting to set.
        """
        return self._scanner

    @scanner.setter
    def scanner(self, scanner):
        if not isinstance(scanner, (RegularGrammar, type(None))):
            raise TypeError('scanner not a RegularGrammar or None')

        self._scanner = scanner

    @property
    def parser(self):
        """Get of set the parser to be used for generation, if any.

        Query for or attempt to set the parser to be used for code generation.
        This allows the child language Generators to be reused rather than
        recreated if attempting to generate source for multiple parsers.

        Args:
          parser (ContextFreeGrammar): The parser to be used for generation. If
              generation of the parser is no longer desired it may also be set
              to None.

        Return:
          None: The default if the parser was not previously set.
          ContextFreeGrammar: If parser was set.

        Raises:
          TypeError: If `parser` is not None or a ContextFreeGrammar when
              attempting to set.
        """
        return self._parser

    @parser.setter
    def parser(self, parser):
        if not isinstance(parser, (ContextFreeGrammar, type(None))):
            raise TypeError('parser not a ContextFreeGrammar or None')

        self._parser = parser

    @property
    def filename(self):
        """Get or set the base filename to be used for generation.

        Query for or attempt to set the base filename to be used for code
        generation of the scanner and/or parser.

        Args:
          filename (str): The base filename to be used for generation.

        Return:
          str: 'out' if not set (default), otherwise the last set filename.

        Raises:
          TypeError if filename is not of type string when being set.
          ValueError if an empty filename is given when being set.
        """
        return self._filename

    @filename.setter
    def filename(self, filename):
        if not isinstance(filename, str):
            raise TypeError('filename must be of type string')

        if not filename:
            raise ValueError('filename must be non empty')

        self._filename = filename

    @property
    def encoding(self):
        """Get or set the encoding to be used for generation.

        Query for or attempt to set the table encoding to be used when
        generating source for the scanner and/or parser. Possible options
        include:
          'direct': encode the table into the source.
          'table': use a small driver and utilize the table for lookup.

        Args:
          encoding (str): the encoding to use when generating code.

        Return:
          str: 'direct' if not set (default), otherwise the last set encoding.

        Raises:
          TypeError if encoding is not if type string
          ValueError if encoding is empty
          ValueError if given encoding is not recognized
        """
        return self._encoding

    @encoding.setter
    def encoding(self, encoding):
        if not isinstance(encoding, str):
            raise TypeError('encoding must be of type string')

        if not encoding:
            raise ValueError('encoding must be non empty')

        if encoding not in ('table', 'direct'):
            raise ValueError('encoding type not recognized')

        self._encoding = encoding

    def _translate(self):
        """The method which subclasses must override to construct the output.

        Override this method in subclasses which should translate the (IR)
        intermediate representation of the given scanner and/or parser to the
        targeted language. It should return the files and there content in a
        dictionary allowing multiple files to be returned for a given language.
        This is the only required function a child generators must implement.

        Return:
          dict[str, str]: Generated file names/paths and associated content.

        Raises:
          NotImplementedError: If not overriden in subclasses.
        """
        raise NotImplementedError('Base Generator incapable of translation')

    def _verify_options(self):
        """Verify the current configuration of set options in the generator.

        Ensure the set options in the given generator do not conflict with one
        another and are a 'sane' configuration for generating output.

        Raises:
          ValueError: if neither a scanner nor a parser is set.
        """
        if not self.scanner and not self.parser:
            raise ValueError('scanner and/or parser must be provided for generation')

    @staticmethod
    def _verify_output(files):
        """Verify the output returned from the subclass generator.

        Ensure the subclasses generated files/content are valid and non-empty.

        Args:
          files (dict[str, str]): Generated file names/paths and content.

        Raises:
          TypeError:  If the generated output is not of type dict.
          ValueError: If no output was generated.
          TypeError:  If any generated filename is not of type string.
          ValueError: If an empty filename was generated.
          TypeError:  If any generated content is not of type string.
          ValueError: If generated file contents are empty.
        """
        if not isinstance(files, dict):
            raise TypeError('generated files must be of type dict')

        if not files:
            raise ValueError('generated files must be of non empty')

        for name, content in files.items():
            if not isinstance(name, str):
                raise TypeError('generated file name must be of type string')

            if not name:
                raise ValueError('generated file name must be non empty')

            if not isinstance(content, str):
                raise TypeError('generated file content must be of type string')

            if not content:
                raise ValueError('generated file content must be non empty')

    def generate(self):
        """Generate a scanner and/or parser with the set generator options.

        Translate the compiled intermediate representation (IR) of the set
        scanner and/or parser to a given output language confined by the choices
        present in options.

        Return:
          dict[str, str]: Generated file names/paths and associated content.

        Raises:
          NotImplementedError: If generation is attempted with base Generator.
          ValueError: if neither a scanner nor a parser is set.
          TypeError:  If the generated output is not of type dict.
          ValueError: If no output was generated.
          TypeError:  If any generated filename is not of type string.
          ValueError: If an empty filename was generated.
          TypeError:  If any generated content is not of type string.
          ValueError: If generated file contents are empty.
        """
        self._verify_options()

        files = self._translate()

        Generator._verify_output(files)

        return files



"""Programmatic creation of parsers through ContextFreeGrammar objects.

The ContextFreeGrammar object represents a BNF (Backus-[Naur/Normal] Form)
grammar which is programmatically transformed into a parse table. This is done
by utilizing the grammars first and follow sets, which are computed internally,
and applying that information to properly construct the resulting parse table.
"""
from copy import deepcopy


class ContextFreeGrammar(object):
    """The ContextFreeGrammar object responsible for creating parse tables.

    ContextFreeGrammar represents a Backus-[Naur/Normal] Form (BNF) language
    grammar which is programmatically transformed into a parse table by finding
    and utilizing the languages first and follow sets. Once created the object
    cannot be mutated, and it will remain static for the rest of it's lifetime.
    All information to properly understand and consume the parse table can be
    queried through the exposed API functions.

    Attributes:
      END_OF_INPUT (int): The marker to use for End Of Input (i.e. '$').
      EPSILON      (int): The marker to use for Epsilon.
    """

    END_OF_INPUT = 0
    EPSILON = 1

    def __init__(self, name, productions, start):
        """Construct a parse table for the given BNF grammar.

        Attempt to initialize a ContextFreeGramamr object with the specified
        name, production rule(s) and start rule identifier. While it is possible
        for any type of grammar to be specified, only left factored LL(1)
        grammars will produce a parse table with a single member per entry. To
        properly specify a LL(1) grammar the following requirements must be met:

          * No left recursion (direct or indirect)
          * Must be left factored
          * No ambiguity

        If any of the above requirements are violated, or the grammar specified
        is not LL(1), there will be a conflict in the table. This means with the
        given set of input productions and with the aid of a single look ahead
        token it cannot be known what rule to choose in order to successfully
        produce a parse without backtracking.

        Args:
          name (str): The name of the input grammar.
          productions (dict[str, list[list[str]]): The production rules of the
              grammar. Dictionary keys are nonterminals and values are a list of
              rules, which are a series of [non]terminal string identifiers. An
              empty rule can be used for epsilon.
          start (str): The start productions nonterminal identifier of the grammar.

        Raises:
          TypeError: if `name` is not a string
          ValueError: if `name` is empty
          TypeError: if `start` is not a string
          ValueError: if `start` is empty
          TypeError: if `productions` is not a dict
          ValueError: if `productions` is empty
          ValueError: if `start` not in `productions`
          TypeError: if any `productions` nonterminal is not a string
          ValueError: if any `productions` nonterminal is empty
          TypeError: if any `productions` nonterminal rules is not a list
          ValueError: if any `productions` nonterminal rules is empty
          TypeError: if any `productions` nonterminal rule is not a list
          TypeError: if any `productions` nonterminal rule contain non strings
          ValueError: if any `productions` nonterminal rule contain empty strings
        """
        if not isinstance(name, str):
            raise TypeError('name must be a string')

        if not name:
            raise ValueError('name must be non empty')

        self._name = name

        if not isinstance(start, str):
            raise TypeError('start must be a string')

        if not start:
            raise ValueError('start must be non empty')

        self._start = start

        if not isinstance(productions, dict):
            raise TypeError('productions must be a dict')

        if not productions:
            raise ValueError('productions must be non empty')

        if self._start not in productions:
            raise ValueError('start production not present in given productions')

        self._rules = []
        for nonterminal, rhs in productions.items():
            if not isinstance(nonterminal, str):
                raise TypeError('production nonterminal must be a string')

            if not nonterminal:
                raise ValueError('production nonterminal must be non empty')

            if not isinstance(rhs, list):
                raise TypeError('production rules must be a list')

            if not rhs:
                raise ValueError('production rules must be non empty')

            for rule in rhs:
                if not isinstance(rule, list):
                    raise TypeError('production rule must be a list')

                for symbol in rule:
                    if not isinstance(symbol, str):
                        raise TypeError('production rule symbol must be a string')

                    if not symbol:
                        raise ValueError('production rule symbol must be non empty')

                self._rules.append((nonterminal, rule))

        self._terminals, self._nonterminals = self._symbols(self._rules)
        self._first_set = self._first(self._terminals, self._nonterminals, self._rules)
        self._follow_set = self._follow(self._nonterminals, self._start,
                                        self._first_set, self._rules)
        self._parse_table, self._rows, self._cols = \
          self._table(self._terminals, self._nonterminals,
                      self._first_set, self._follow_set, self._rules)

    @property
    def name(self):
        """Query for the name of the grammar.

        A readonly property which copies the grammar's name to protect against
        user mutations.

        Return:
          str: The given input name of the grammar.
        """
        return deepcopy(self._name)

    @property
    def start(self):
        """Query for the start production's nonterminal of the grammar.

        A readonly property which copies the grammar's start production
        nonterminal to protect against user mutations.

        Return:
          str: The given start production nonterminal.
        """
        return deepcopy(self._start)

    @property
    def terminals(self):
        """Query for the terminal set of the grammar.

        A readonly property which copies the grammar's terminal set to protect
        against user mutations.

        Return:
          set[str]: The set of terminal symbols contained in the grammar.
        """
        return deepcopy(self._terminals)

    @property
    def nonterminals(self):
        """Query for the nonterminal set of the grammar.

        A readonly property which copies the grammar's nonterminal set to
        protect against user mutations.

        Return:
          set[str]: The set of nonterminal production symbols in the grammar.
        """
        return deepcopy(self._nonterminals)

    @property
    def first(self):
        """Query for the first set(s) of the grammar's [non]terminal(s).

        A readonly property which copies the first set(s) of the grammar's
        [non]terminal(s) to protect against user mutations.

        Return:
          dict[str, set[str, int]]: The first set of every [non]terminal.
        """
        return deepcopy(self._first_set)

    @property
    def follow(self):
        """Query for the follow set(s) of the grammar's nonterminal(s).

        A readonly property which copies the follow set(s) grammar's
        nonterminals to protect against user mutations.

        Return:
          dict[str, set[str, int]]: The follow set of every nonterminal.
        """
        return deepcopy(self._follow_set)

    @property
    def rules(self):
        """Query for the production rules of the grammar.

        A readonly property which copies the grammar's flattened production
        rules to protect against user mutations.

        Return:
          list[tuple[str, list[str]]]: A flattened list of production rules.
        """
        return deepcopy(self._rules)

    @property
    def table(self):
        """Query for the parse table of the given input grammar.

        A readonly property which copies the grammar's parse table to protect
        against user mutations.

        Return:
          list[list[set[int]]]: row-major parse table with list rule
            inidice(s) entries. (i.e. Table[nonterminal][terminal] -> rules)
          dict[str, int]: Mapping for row (nonterminal) symbol to table index.
          dict[str, int]: Mapping for column (terminal) symbol to table index.
        """
        return deepcopy(self._parse_table), \
               deepcopy(self._rows), \
               deepcopy(self._cols)

    @staticmethod
    def _symbols(productions):
        """Collect the [non]terminals from the given input productions.

        Report all literal terminal symbols and non terminal production symbols
        appearing in the productions present.

        Args:
          productions (list[tuple[str, list[str]]]): The input production rules.

        Return:
          set[str]: The set of terminal symbols contained in the grammar.
          set[str]: The set of nonterminal production symbols in the grammar.
        """
        terminals, nonterminals = set(), set()
        for nonterminal, rule in productions:
            terminals.update(rule)
            nonterminals.add(nonterminal)
        terminals = terminals - nonterminals
        return terminals, nonterminals

    @staticmethod
    def _first_production(production, first):
        """Derive the first set given a single production rule.

        Compute the first set for a single nonterminal's production rule
        following the algorithm at:
        http://marvin.cs.uidaho.edu/Teaching/CS445/firstfollow.txt

        Args:
          production (list[str]): a series of [non]terminal symbols.
          first (dict[str, set[str, int]]): first sets for [non]terminals.

        Return:
          set[str, int]: possible symbols first encountered for `production`.
        """
        _first = set([ContextFreeGrammar.EPSILON])
        for symbol in production:
            _first.update(first[symbol])
            if ContextFreeGrammar.EPSILON not in first[symbol]:
                _first.discard(ContextFreeGrammar.EPSILON)
                break
        return _first

    @staticmethod
    def _first(terminals, nonterminals, productions):
        """Derive the grammars first sets.

        Calculate the first set for each [non]terminal in the grammar following
        the algorithm at:
        http://marvin.cs.uidaho.edu/Teaching/CS445/firstfollow.txt

        Args:
          terminals (set[str]): set of grammar terminal symbols.
          nonterminals (set[str]): set of grammar nonterminal symbols.
          productions (list[tuple[str, list[str]]]): flattened list of rules.

        Return:
          dict[str, set[str, int]]: The first set of every [non]terminal.
        """
        first = {}

        # foreach elem A of TERMS do first[A] = {A}
        for terminal in terminals:
            first[terminal] = set([terminal])

        # foreach elem A of NONTERMS do first[A] = {}
        for nonterminal in nonterminals:
            first[nonterminal] = set()

        # loop until nothing new happens updating the first sets
        while True:
            changed = False

            for nonterminal, production in productions:
                new = ContextFreeGrammar._first_production(production, first) - first[nonterminal]
                if new:
                    first[nonterminal].update(new)
                    changed = True

            if not changed:
                return first

    @staticmethod
    def _follow(nonterminals, start, first, productions):
        """Derive the grammars follow sets.

        Calculate the follow set for each nonterminal in the grammar following
        the algorithm at:
        http://marvin.cs.uidaho.edu/Teaching/CS445/firstfollow.txt

        Args:
          nonterminals (set[str]): set of grammar nonterminal symbols.
          start (str): The given start production nonterminal.
          first (dict[str, set[str, int]]): first sets for [non]terminals.
          productions (list[tuple[str, list[str]]]): flattened list of rules.

        Return:
          dict[str, set[str, int]]: The follow set of every nonterminal.
        """
        # foreach elem A of NONTERMS do follow[A] = {}
        follow = {nonterminal: set() for nonterminal in nonterminals}

        # put $ (end of input marker) in follow(<Start>)
        follow[start] = set([ContextFreeGrammar.END_OF_INPUT])

        # loop until nothing new happens updating the follow sets
        while True:
            changed = False

            for nonterminal, production in productions:
                for (idx, elem) in enumerate(production):
                    if elem in nonterminals:
                        new = ContextFreeGrammar._first_production(production[idx+1:], first)
                        if ContextFreeGrammar.EPSILON in new:
                            new.discard(ContextFreeGrammar.EPSILON)
                            new.update(follow[nonterminal])

                        new = new - follow[elem]
                        if new:
                            follow[elem].update(new)
                            changed = True

            if not changed:
                return follow

    @staticmethod
    def _table(terminals, nonterminals, first, follow, productions):
        """Programmatically construct the grammars parse table.

        Construct the parse table indexed by nonterminal x terminal. This is
        done by finding the predict sets, derived from the first sets,
        following the algorithm at:
        http://marvin.cs.uidaho.edu/Teaching/CS445/firstfollow.txt

        Args:
          terminals (set[str]): set of grammar terminal symbols.
          nonterminals (set[str]): set of grammar nonterminal symbols.
          first (dict[str, set[str, int]]): first sets for [non]terminals.
          follow (dict[str, set[str, int]]): follow sets for terminals.
          productions (list[tuple[str, list[str]]]): flattened list of rules.

        Return:
          list[list[set[int]]]: row-major parse table with list rule
            inidice(s) entries. (i.e. Table[nonterminal][terminal] -> rules)
          dict[str, int]: Mapping for row (nonterminal) symbol to table index.
          dict[str, int]: Mapping for column (terminal) symbol to table index.
        """
        rows = {n:i for i, n in enumerate(nonterminals)}
        cols = {t:i for i, t in enumerate(terminals | set([ContextFreeGrammar.END_OF_INPUT]))}

        table = [[set() for _ in cols] for _ in rows]

        for (rule, (nonterminal, production)) in enumerate(productions):
            predict = ContextFreeGrammar._first_production(production, first)
            if ContextFreeGrammar.EPSILON in predict:
                predict.discard(ContextFreeGrammar.EPSILON)
                predict.update(follow[nonterminal])
            for terminal in predict:
                table[rows[nonterminal]][cols[terminal]].add(rule)

        return table, rows, cols



"""Python generator.

A scanner/parser generator targeting python. Generates a single python (.py)
file.
"""
from spag.generator import Generator


class Python(Generator):
    """Python generator.

    A simple object for compiling scanner's and/or parser's to python programs.
    """

    def _translate(self):
        """Override the superclass method to generate source code.

        Attempt to generate the python (.py) source file with the corresponding
        scanner and/or parser currently set in the object.
        """
        return {self.filename+'.py': super(Python, self)._translate()}



"""SPaG package management.

Create the PyPI package distribution(s) for use with python pip or install
directly from source.
"""
from setuptools import setup


with open('README.md', 'r') as fd:
    README = fd.read()

setup(
    name='SPaG',
    version='1.0.0a0',
    license='MIT',
    author='Ryan Rozanski',
    author_email='',
    maintainer='Ryan Rozanski',
    maintainer_email='',
    description=(
        'A module containing scanner (regular expression) and parser (BNF) '
        'compilers as well as a base generator, which provides protection and '
        'validation, from which all target language generators must inherit '
        'from. A script is also included which reads the respective '
        'specification(s) from file and outputs the resulting code to disk.'
    ),
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/rrozansk/SPaG',
    download_url='https://github.com/rrozansk/SPaG',
    python_requires='>= 2.7.0',
    keywords=' '.join([
        'abstract-syntax-tree',
        'ast',
        'AST',
        'Backus-Naur-form',
        'Backus-normal-form',
        'bnf',
        'BNF',
        'command-line-interface',
        'cli',
        'CLI',
        'code-generation',
        'compiler',
        'context-free-grammar',
        'deterministic-finite-automaton',
        'dfa',
        'DFA',
        'dfa-minimization',
        'direct-encoding',
        'finite-state-machine',
        'first-set',
        'follow-set',
        'fsm',
        'FSM',
        'lexeme',
        'lexer',
        'lexer-generator',
        'lexical-analysis',
        'll1',
        'LL1',
        'LL(1)',
        'longest-match',
        'maximal-munch',
        'nfa',
        'NFA',
        'nondeterministic-finite-automaton',
        'nonterminal(s)',
        'parser',
        'parser-generator',
        'production-rule(s)',
        'regular-expression(s)',
        'regular-grammar',
        'scanner',
        'scanner-generator',
        'script',
        'state-transition-table',
        'table-encoding',
        'terminal(s)',
        'tokenization',
        'tokenizer',
        'tokens'
    ]),
    packages=[
        'spag',
        'spag.generators'
    ],
    package_data={
        'spag': [
            '../LICENSE.txt'
        ]
    },
    scripts=[
        'scripts/spag_cli'
    ],
    include_package_data=True,
    zip_safe=True,
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Compilers',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Linguistic'
    )
)
